"""command-line interface for messages package."""

import os
import sys

import click
from click import argument
from click import option


from messages import MESSAGES
from .api import send
from .exceptions import UnsupportedMessageTypeError


def check_args(ctx, kwds):
    """If no CLI args given, prints usage and exits."""
    if not any([val for key, val in kwds.items()]):
        click.echo(ctx.get_usage())
        click.echo('Try messages --help for more information.')
        sys.exit(0)


def check_type(kwds):
    """If no or incorrect message-type specified, alerts user and exits."""
    if not kwds['type']:
        click.echo('[!] Must specify message type with -t option.')
        click.echo('Try messages --help for more information.')
        sys.exit(0)
    if kwds['type'].lower() not in MESSAGES:
        raise UnsupportedMessageTypeError(kwds['type'])


def get_body(kwds):
    """Reads message body if specified via filepath."""
    if kwds['file'] and os.path.isfile(kwds['file']):
        kwds['body'] = open(kwds['file'], 'r').read()
        kwds['file'] = None


def trim_args(kwds):
    """Gets rid of args with value of None, as well as select keys."""
    reject_key = ('type')
    reject_val = (None, ())
    kwargs = {k:v for k,v in kwds.items() if k not in ('type') and
                                             v not in (None, ())}
    for k, v in kwargs.items():
        if k in ('to', 'cc', 'bcc', 'attach'):
            kwargs[k] = list(kwargs[k])
    return kwargs


@click.command()
@argument('type', required=True)
@option('-f', '--from',
    help='From address/phone/etc.')
@option('-t', '--to', multiple=True,
    help='Primary (To) recipient.')
@option('-c', '--carboncopy', multiple=True,
    help='Carbon Copy (CC) addresses.')
@option('-b', '--blindcopy', multiple=True,
    help='Blind Carbon Copy (BCC) addresses.')
@option('-s', '--subject',
    help='Subject line.')
@option('-m', '--body',
    help='Message body text.')
@option('-M', '--file',
    help='Read message body from filepath.')
@option('-a', '--attach', multiple=True,
    help='Attachments -- filepath or url to attach.')
@option('-S', '--save',
    help='Save default values/credentials.')
@option('-P', '--profile',
    help='Specify pre-configured user profile.')
@option('-T', '--types',
    help='List available message types.')
@click.version_option(version='0.3.1', prog_name='Messages')
@click.pass_context
def main(ctx, **kwds):
    """Specify Message-Type, Recipients, and Content to send."""

    check_args(ctx, kwds)
    check_type(kwds)
    get_body(kwds)
    kwargs = trim_args(kwds)

    send(kwds['type'], send_async=True, **kwargs)



"""messages.exceptions tests."""

import pytest

from messages.exceptions import InvalidMessageInputError
from messages.exceptions import UnsupportedMessageTypeError
from messages.exceptions import UnknownProfileError


##############################################################################
# TESTS: MessageInputError
##############################################################################

def test_InvalidMessageInputError(capsys):
    """
    GIVEN no object instantiated, just raise the exception
    WHEN the exception is raised with given args
    THEN assert it raises exception and prints proper output
    """
    with pytest.raises(InvalidMessageInputError):
        raise InvalidMessageInputError('Email', 'from_', 'email address')
        out, err = capsys.readouterr()
        expected = 'Invalid input for specified message class: Email'
        expected += '\n\t* argument: "from_"'
        expected += '\n\t* input type must be: email address\n'
        assert out == expected
        assert err == ''


def test_UnsupportedMessageTypeError_default(capsys):
    """
    GIVEN no object instantiated, just raise the exception
    WHEN the exception is raised with given args
    THEN assert it raises exception and prints proper output
    """
    with pytest.raises(UnsupportedMessageTypeError):
        raise UnsupportedMessageTypeError('BadType')
        out, err = capsys.readouterr()
        expected = 'Invalid input for specified message class: BadType\n'
        assert out == expected
        assert err == ''


def test_UnsupportedMessageTypeError_listarg(capsys):
    """
    GIVEN no object instantiated, just raise the exception
    WHEN the exception is raised with given args
    THEN assert it raises exception and prints proper output
    """
    with pytest.raises(UnsupportedMessageTypeError):
        raise UnsupportedMessageTypeError('BadType', {'m1', 'm2'})
        out, err = capsys.readouterr()
        expected = 'Invalid input for specified message class: BadType'
        expected += '\n\t* Supported message types: '
        expected += "{'m1', 'm2'}\n"
        assert out == expected
        assert err == ''


def test_UnknownProfileError(capsys):
    """
    GIVEN no object instantiated, just raise the exception
    WHEN the exception is raised with given args
    THEN assert it raises exception and prints proper output
    """
    with pytest.raises(UnknownProfileError):
        raise UnknownProfileError('unknown_user')
        out, err = capsys.readouterr()
        expected = 'Unknown Profile name: unknown_user\n'
        assert out == expected
        assert err == ''



"""messages.email_ tests."""

import getpass
import os
import pytest
import smtplib

from collections import deque
from email.mime.multipart import MIMEMultipart
from unittest.mock import patch

import messages
from messages.email_ import Email
from messages.email_ import configure
from messages._eventloop import MESSAGELOOP


##############################################################################
# FIXTURES
##############################################################################

@pytest.fixture()
@patch.object(messages.email_, 'configure')
def get_email(config_mock):
    """Return a valid Email object."""
    e = Email(from_='me@here.com', to='you@there.com',
            server='smtp.gmail.com', port=465, password='password',
            cc='someone@there.com', bcc='them@there.com',
            subject='subject', body='message', attachments=['file1', 'file2'],
            profile='myName', save=False)
    e.from_ = 'me@here.com'
    e.server = 'smtp.gmail.com'
    e.port = 465
    e.password = 'password'
    return e


# skip this test if on travs-ci
travis = pytest.mark.skipif("TRAVIS" in os.environ and
                    os.environ["TRAVIS"] == "true",
                    reason='skipping test if on travis-ci')


# skip this test if NOT on travis-ci
not_travis = pytest.mark.skipif("TRAVIS" not in os.environ,
                    reason='skipping test if not on travis-ci')


##############################################################################
# TESTS: Email.__init__
##############################################################################

def test_email_init_normal(get_email):
    """
    GIVEN a need to create an Email object
    WHEN the user instantiates a new object with required args
    THEN assert Email object is created with given args
    """
    e = get_email
    assert e is not None
    assert e.server == 'smtp.gmail.com'
    assert e.port == 465
    assert e.password == 'password'
    assert e.from_ == 'me@here.com'
    assert e.to == 'you@there.com'
    assert e.cc == 'someone@there.com'
    assert e.bcc == 'them@there.com'
    assert e.subject == 'subject'
    assert e.body == 'message'
    assert e.attachments == ['file1', 'file2']
    assert e.message is None
    assert isinstance(e.sent_messages, deque)


##############################################################################
# TESTS: Email.__repr__
##############################################################################

def test_email_repr(get_email, capsys):
    """
    GIVEN a valid Email object
    WHEN the user calls repr() or `>>> e`
    THEN assert the correct format prints
    """
    e = get_email
    print(repr(e))
    out, err = capsys.readouterr()
    assert '<messages.Email class> at: ' in out
    assert err == ''


##############################################################################
# TESTS: Email.__str__
##############################################################################

def test_email_str(get_email, capsys):
    """
    GIVEN a valid Email object
    WHEN the user calls print() on the Email object
    THEN assert the correct format prints
    """
    e = get_email
    expected = ('MIMEMultipart Email:'
                '\n\tServer: smtp.gmail.com:465'
                '\n\tFrom: me@here.com'
                '\n\tTo: you@there.com'
                '\n\tCc: someone@there.com'
                '\n\tBcc: them@there.com'
                '\n\tSubject: subject'
                '\n\tbody: message...'
                '\n\tattachments: [\'file1\', \'file2\']\n')
    print(e)
    out, err = capsys.readouterr()
    assert out == expected
    assert err == ''


##############################################################################
# TESTS: Email.get_server
##############################################################################

def test_get_server_known(get_email):
    """
    GIVEN a valid Email object
    WHEN get_server() is called with an email address
    THEN assert a known smtp server is returned
    """
    e = get_email
    server = e.get_server('me@gmail.com')
    assert server == ('smtp.gmail.com', 465)


def test_get_server_guess(get_email):
    """
    GIVEN a valid Email object
    WHEN get_server() is called with an email address
    THEN assert an smtp server guess is returned
    """
    e = get_email
    server = e.get_server('me@test.com')
    assert server == ('smtp.test.com', 465)


##############################################################################
# TESTS: Email.list_to_string
##############################################################################

def test_list_to_string(get_email):
    """
    GIVEN a valid Email object
    WHEN given a list of recipients
    THEN assert the list is returned as a single string
    """
    e = get_email
    output = e.list_to_string(['her@there.com', 'him@there.com'])
    assert output == 'her@there.com, him@there.com'


##############################################################################
# TESTS: Email.generate_email
##############################################################################

@patch.object(Email, 'add_attachments')
@patch.object(Email, 'add_body')
@patch.object(Email, 'add_header')
def test_generate_email(header_mock, body_mock, attach_mock, get_email):
    """
    GIVEN a valid Email object
    WHEN Email.generate_email() is called
    THEN assert the email structure is created
    """
    e = get_email
    e.generate_email()
    assert isinstance(e.message, MIMEMultipart)
    assert header_mock.call_count == 1
    assert body_mock.call_count == 1
    assert attach_mock.call_count == 1


##############################################################################
# TESTS: Email.add_header
##############################################################################

@patch.object(Email, 'add_attachments')
@patch.object(Email, 'add_body')
def test_add_header(body_mock, attach_mock, get_email):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
    WHEN Email.add_header() is called
    THEN assert correct parameters are set
    """
    e = get_email
    e.generate_email()
    assert e.message['From'] == 'me@here.com'
    assert e.message['Subject'] == 'subject'


##############################################################################
# TESTS: Email.add_body
##############################################################################

@patch.object(MIMEMultipart, 'attach')
@patch.object(Email, 'add_attachments')
@patch.object(Email, 'add_header')
def test_add_body(header_mock, attach_mock, mime_attach_mock, get_email):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
    WHEN Email.add_body() is called
    THEN assert body_text is attached
    """
    e = get_email
    e.generate_email()
    assert mime_attach_mock.call_count == 1


##############################################################################
# TESTS: Email.add_attachments
##############################################################################

@travis
@patch.object(MIMEMultipart, 'attach')
@patch.object(Email, 'add_header')
@patch.object(Email, 'add_body')
def test_add_attachments_list_local(body_mock, header_mock, mime_attach_mock,
                               get_email):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
        and Email.attachments is a list
    WHEN Email.add_attachments() is called
    THEN assert correct attachments are attached
    """
    e = get_email
    e.attachments = ['./data/file1.txt', './data/file2.png',
                     './data/file3.pdf', './data/file4.xlsx']
    e.generate_email()
    assert mime_attach_mock.call_count == 4


@not_travis
@patch.object(MIMEMultipart, 'attach')
@patch.object(Email, 'add_header')
@patch.object(Email, 'add_body')
def test_add_attachments_list_travis(body_mock, header_mock, mime_attach_mock,
                                get_email):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
         and Email.attachments is a list
    WHEN Email.add_attachments() is called
    THEN assert correct attachments are attached
    """
    e = get_email
    PATH = '/home/travis/build/trp07/messages/tests/data/'
    e.attachments = [PATH + 'file1.txt', PATH + 'file2.png',
                     PATH + 'file3.pdf', PATH + 'file4.xlsx']
    e.generate_email()
    assert mime_attach_mock.call_count == 4


@travis
@patch.object(MIMEMultipart, 'attach')
@patch.object(Email, 'add_header')
@patch.object(Email, 'add_body')
def test_add_attachments_str_local(body_mock, header_mock, mime_attach_mock,
                               get_email):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
        and Email.attachments is a str
    WHEN Email.add_attachments() is called
    THEN assert correct attachments are attached
    """
    e = get_email
    e.attachments = './data/file1.txt'
    e.generate_email()
    assert mime_attach_mock.call_count == 1


@not_travis
@patch.object(MIMEMultipart, 'attach')
@patch.object(Email, 'add_header')
@patch.object(Email, 'add_body')
def test_add_attachments_str_travis(body_mock, header_mock, mime_attach_mock,
                                get_email):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
         and Email.attachments is a str
    WHEN Email.add_attachments() is called
    THEN assert correct attachments are attached
    """
    e = get_email
    PATH = '/home/travis/build/trp07/messages/tests/data/'
    e.attachments = PATH + 'file1.txt'
    e.generate_email()
    assert mime_attach_mock.call_count == 1


##############################################################################
# TESTS: Email.get_session
##############################################################################

@patch.object(Email, 'get_ssl')
@patch.object(Email, 'generate_email')
def test_get_session_ssl(gen_email_mock, getssl_mock, get_email):
    """
    GIVEN a valid Email object
    WHEN Email.get_session() is called
    THEN assert the correct functions are called
    """
    e = get_email
    e.generate_email()
    e.get_session()
    assert getssl_mock.call_count == 1


@patch.object(Email, 'get_tls')
@patch.object(Email, 'generate_email')
def test_get_session_tls(gen_email_mock, gettls_mock, get_email):
    """
    GIVEN a valid Email object
    WHEN Email.get_session() is called
    THEN assert the correct functions are called
    """
    e = get_email
    e.port = 587
    e.generate_email()
    e.get_session()
    assert gettls_mock.call_count == 1


##############################################################################
# TESTS: Email.get_ssl
##############################################################################

@patch.object(smtplib, 'SMTP_SSL')
def test_get_ssl(smtpssl_mock, get_email):
    """
    GIVEN a valid Email object
    WHEN Email.get_ssl() is called
    THEN assert an SMTP_SSL instance is invoked
    """
    e = get_email
    e.get_ssl()
    assert smtpssl_mock.call_count == 1


##############################################################################
# TESTS: Email.get_tls
##############################################################################

@patch.object(smtplib, 'SMTP')
def test_get_tls(smtp_mock, get_email):
    """
    GIVEN a valid Email object
    WHEN Email.get_tls() is called
    THEN assert an SMTP instance is invoked
    """
    e = get_email
    e.get_tls()
    assert smtp_mock.call_count == 1


##############################################################################
# TESTS: Email.send
##############################################################################

@patch.object(Email, 'get_session')
@patch.object(Email, 'add_attachments')
@patch.object(Email, 'add_body')
@patch.object(Email, 'add_header')
def test_send(header_mock, body_mock, attach_mock, session_mock,
              get_email, capsys):
    """
    GIVEN a valid Email object
    WHEN Email.send() is called
    THEN assert the correct functions are called and correct attributes
        updated
    """
    e = get_email
    e.send()
    out, err = capsys.readouterr()
    assert out == 'Message sent...\n'
    assert err == ''
    assert e.sent_messages[0] == repr(e)


##############################################################################
# TESTS: Email.send_async
##############################################################################

@patch.object(MESSAGELOOP, 'add_message')
def test_send_async(msg_loop_mock, get_email):
    """
    GIVEN a valid Email object
    WHEN Email.send_async() is called
    THEN assert it is added to the event loop for async sending
    """
    e = get_email
    e.send_async()
    assert msg_loop_mock.call_count == 1



"""
Configuration module.

Handles the creation/editing of configuration files and saving of
default attributes, credentials, etc.
"""

from getpass import getpass

import jsonconfig

from .exceptions import UnknownProfileError


def configure(msg, params, to_save, credentials):
    """
    Get default values and credentials from a configuration profile.
    If not created, and save=True in the msg class parameter, then
    configuration file is created.

    Args:
        :msg: (Message class) an instance of a message class
        :params: (dict) dict of params passed to message classes' __init__
            method
        :to_save: (set or list) list of default values to save (from address,
            server name, etc.)
        :credentials: (set or list) list of credentials (password, auth_tokens,
            etc.)
    """
    msg_type = msg.__class__.__name__.lower()
    defaults = set(to_save)
    creds = set(credentials)

    with jsonconfig.Config('messages') as cfg:
        if params['profile'] is None:
            msg.profile = 'default_user'
        else:
            msg.profile = params['profile']

        if msg.profile not in cfg.data:
            cfg.data[msg.profile] = {}

        if msg_type not in cfg.data[msg.profile]:
                cfg.data[msg.profile][msg_type] = {}

        for d in defaults:
            setattr(msg, d, cfg.data[msg.profile][msg_type].get(d, params[d]))

        for c in creds:
            setattr(msg, c, (params[c] or cfg.pwd.get(
                msg.profile + '_' + msg_type, None)))
            if getattr(msg, c) is None:
                setattr(msg, c, getpass('\n' + c + ': '))

        if params['save']:
            for d in defaults:
                cfg.data[msg.profile][msg_type][d] = getattr(msg, d)
            for c in creds:
                cfg.pwd[(msg.profile + '_' + msg_type)] = getattr(msg, c)
        cfg.kwargs['dump']['indent'] = 4


def set_default_profile(profile):
    """
    Sets the default profile to use in the messages config.json file.

    Args:
        :profile: (str) name of existing profile to use
    """
    with jsonconfig.Config('messages') as cfg:
        if profile not in cfg.data:
            raise UnknownProfileError(profile)
        cfg.data['default'] = profile
        cfg.kwargs['dump']['indent'] = 4




"""messages.config tests."""

import pytest
from unittest.mock import patch

import jsonconfig

import messages.config
from messages.config import configure
from messages.config import set_default_profile
from messages.config import getpass
from messages.exceptions import UnknownProfileError


##############################################################################
# FIXTURES
##############################################################################

class Msg:
    """A test message class."""
    def __init__(self, from_, password, profile=None, save=False):
        self.config_kwargs = {'from_': from_, 'password': password,
                     'profile': profile, 'save': save}


@pytest.fixture()
def get_msg():
    return Msg(from_='me', password='passw0rd')


##############################################################################
# TESTS: configure
##############################################################################

def test_configure_profileNone_saveFalse(get_msg, cfg_mock):
    """
    GIVEN a valid message object
    WHEN `configure` is called with the specified args
    THEN assert appropriate attributes are set
    """
    msg = get_msg
    configure(msg, params=msg.config_kwargs, to_save={'from_'}, credentials={'password'})
    assert msg.from_ == 'me'
    assert msg.password == 'passw0rd'


def test_configure_profileYes_saveTrue(cfg_mock):
    """
    GIVEN a valid message object
    WHEN `configure` is called with the specified args
    THEN assert appropriate attributes are set
    """
    msg = Msg(from_='me', password='passw0rd', profile='myProf', save=True)
    configure(msg, params=msg.config_kwargs, to_save={'from_'}, credentials={'password'})
    assert msg.from_ == 'me'
    assert msg.password == 'passw0rd'


@patch.object(messages.config, 'getpass')
def test_configure_noPassword(getpass_mock, cfg_mock):
    """
    GIVEN a valid message object
    WHEN `configure` is called with the specified args
    THEN assert appropriate attributes are set
    """
    msg = Msg(from_='me', password=None, profile='NewProf', save=False)
    configure(msg, params=msg.config_kwargs, to_save={'from_'}, credentials={'password'})
    print(msg.password)
    assert getpass_mock.call_count == 1


##############################################################################
# TESTS: set_default_profile
##############################################################################

def test_set_default_profileGood(cfg_mock):
    """
    GIVEN a profile to make as the default
    WHEN set_default_profile(profile) is called
    THEN verify no errors occur with call to jsonconfig.Config
    """
    set_default_profile('tester')


def test_set_default_profileBad(cfg_mock):
    """
    GIVEN a profile to make as the default
    WHEN set_default_profile(profile) is called with an unknown profile name
    THEN assert UnknownProfileError is raised
    """
    with pytest.raises(UnknownProfileError):
        set_default_profile('UnknownProfile')



"""messages._utils tests."""

import pytest
from unittest.mock import patch

from messages.exceptions import InvalidMessageInputError
from messages._utils import Validator
from messages._utils import validus


##############################################################################
# FIXTURES
##############################################################################

@pytest.fixture()
def get_validator():
    """Return a valid Validator object."""
    return Validator()


class Email:
    """Basic Email class used for testing."""
    def __init__(self, from_='me@here.com'):
        self.from_ = from_


@pytest.fixture()
def get_testemail():
    """Return a valid ForTest object for testing."""
    return Email()


##############################################################################
# TEST: Validator.validate_input
##############################################################################

@patch.object(Validator, '_is_valid')
def test_val_input_singular(isval_mock, get_validator, get_testemail):
    """
    GIVEN a valid Validator object
    WHEN validate_input() is called on a message object
        whose attribute is a singular valid input
    THEN assert _is_valid is called upon the correct attribute
    """
    v = get_validator
    e = get_testemail
    v.validate_input(e, 'from_')
    assert isval_mock.call_count == 1


@patch.object(Validator, '_is_valid')
def test_val_input_list(isval_mock, get_validator, get_testemail):
    """
    GIVEN a valid Validator object
    WHEN validate_input() is called on a message object
        whose attribute is a list of valid inputs
    THEN assert _is_valid is called upon all items in the attribute
    """
    v = get_validator
    e = get_testemail
    e.from_ = ['me@here.com', 'you@there.com']
    v.validate_input(e, 'from_')
    assert isval_mock.call_count == 2


@patch.object(Validator, '_is_valid')
def test_val_input_singular_raiseError(isval_mock, get_validator,
    get_testemail):
    """
    GIVEN a valid Validator object
    WHEN validate_input() is called on a message object
        whose attribute is a singular invalid input
    THEN assert InvalidMessageInputError is raised
    """
    v = get_validator
    e = get_testemail
    e.from_ = 'bademail'
    isval_mock.return_value = False
    with pytest.raises(InvalidMessageInputError):
        v.validate_input(e, 'from_')


@patch.object(Validator, '_is_valid')
def test_val_input_list_raiseError(isval_mock, get_validator,
    get_testemail):
    """
    GIVEN a valid Validator object
    WHEN validate_input() is called on a message object
        whose attribute is a list of invalid inputs
    THEN assert InvalidMessageInputError is raised
    """
    v = get_validator
    e = get_testemail
    e.from_ = ['bademail1', 'bademail2']
    isval_mock.return_value = False
    with pytest.raises(InvalidMessageInputError):
        v.validate_input(e, 'from_')


##############################################################################
# TEST: Validator._is_valid
##############################################################################

@patch.object(validus, 'isemail')
def test_isvalid_true(isemail_mock, get_validator, get_testemail):
    """
    GIVEN a valid Validator object
    WHEN when _is_valid() is called with upon a valid attribute
    THEN assert True is returned
    """
    v = get_validator
    e = get_testemail
    isemail_mock.return_value = True
    result = v._is_valid('EMAIL', 'from_', 'you@there.com')
    assert result == True


@patch.object(validus, 'isemail')
def test_isvalid_false(isemail_mock, get_validator, get_testemail):
    """
    GIVEN a valid Validator object
    WHEN when _is_valid() is called with upon an invalid attribute
    THEN assert False is returned
    """
    v = get_validator
    e = get_testemail
    isemail_mock.return_value = False
    result = v._is_valid('EMAIL', 'from_', 'bademail')
    assert result == False



"""
Module designed to make creating and sending text messages easy.

1.  Twilio
    - Uses the Twilio API to send text messages.
    - Must have an account_sid, auth_token, and a twilio phone number
      in order to use.
    - Go to https://www.twilio.com to register.
"""

import sys
from collections import deque

from twilio.rest import Client

from .config import configure
from ._eventloop import MESSAGELOOP
from ._interface import Message


class Twilio(Message):
    """
    Create and send text SMS/MMS text messages using the Twilio API.

    Args:
        :from_: (str) phone number of originating text, e.g. '+15558675309'
        :to: (str) phone number of destination text, e.g. '+15558675309'
        :acct_sid: (str) api credential from twilio
        :auth_token: (str) api credential from twilio
        :body: (str) message to send
        :attachments: (str) url of any image to send along with message
        :profile: (str) use a separate account profile specified by name
        :save: (bool) save pertinent values in the messages config file,
            such as from_, acct_sid, auth_token (encrypted keyring) to make
            sending messages faster.

    Attributes:
        :client: (Client) twilio.rest client for authentication
        :sid: (str) return value from send, record of sent message

    Usage:
        Create a text message (SMS/MMS) object with required Args above.
        Send text message with self.send() or self.send_async() methods.

    Notes:
        For API description:
        https://www.twilio.com/docs/api/messaging/send-messages
    """

    def __init__(
        self, from_=None, to=None, acct_sid=None, auth_token=None,
        body='', attachments=None, profile=None, save=False
    ):

        config_kwargs = {'from_': from_, 'acct_sid': acct_sid,
                'auth_token': auth_token, 'profile': profile, 'save': save}

        configure(self, params=config_kwargs,
                to_save={'from_', 'acct_sid'}, credentials={'auth_token'})

        self.to = to
        self.body = body
        self.attachments = attachments
        self.client = None
        self.sid = None
        self.sent_messages = deque()


    def __str__(self):
        """print(Twilio(**args)) method."""
        return('Twilio Text Message:'
               '\n\tFrom: {}'
               '\n\tTo: {}'
               '\n\tbody: {}...'
               '\n\tattachments: {}'
               .format(self.from_, self.to, self.body, self.attachments))


    def get_client(self):
        """Return a twilio.rest client."""
        return Client(self.acct_sid, self.auth_token)

    def send(self):
        """
        Send the SMS/MMS message.
        Set self.sid to return code of message.
        Append the (sid, message) tuple to self.sent_messages as a history.
        """
        if self.client is None:
            self.client = self.get_client()
        msg = self.client.messages.create(
              to=self.to,
              from_=self.from_,
              body=self.body,
              media_url=self.attachments,
            )
        self.sid = msg.sid
        print('Message sent...', file=sys.stdout)
        self.sent_messages.append((msg.sid, repr(self)))


    def send_async(self):
        """Send message asynchronously."""
        MESSAGELOOP.add_message(self)



"""
Module designed to make creating and sending chat messages easy.

1.  SlackWebhook
    - Send messages via the Incoming Webhooks feature
    - https://api.slack.com/incoming-webhooks
"""

from collections import deque

import requests

from .config import configure
from ._eventloop import MESSAGELOOP
from ._interface import Message


class SlackWebhook(Message):
    """
    Create and send Slack message via the Incoming WebHooks API.

    Args:
        :from_: (str) optional arg to specify who message is from.
        :url: (str) webhook url for installed slack app.
        :subject: (str) optional arg to specify message subject.
        :body: (str) message to send.
        :attachments: (str or list) each item is a url to attach
        :params: (dict) additional attributes to add to each attachment,
            i.e. author_name, title, text, etc., see API for information
            on which attributes are possible.
        :profile: (str) use a separate account profile specified by name
        :save: (bool) save pertinent values in the messages config file,
            such as from_, url (encrypted keyring) to make
            sending messages faster.

    Attributes:
        :message: (dict) current form of the message to be constructed
        :sent_messages: (deque) all messages sent with current SlackWebhook
            object, acting as a log of messages sent in the current session.

    Usage:
        Create a SlackWebhook object with required Args above.
        Send message with self.send() or self.send_async() methods.

    Note:
        For API description:
        https://api.slack.com/incoming-webhooks
    """

    def __init__(
        self, from_=None, url=None, subject=None, body='', attachments=None,
        params=None, profile=None, save=False
    ):

        config_kwargs = {'from_': from_, 'url': url, 'profile': profile,
                'save': save}

        configure(self, params=config_kwargs,
                to_save={'from_', 'url'}, credentials={})

        self.subject = subject
        self.body = body
        self.attachments = attachments
        self.params = params
        self.message = {}
        self.sent_messages = deque()


    def construct_message(self):
        """Build the message params."""
        self.message['text'] = ''
        if self.from_:
            self.message['text'] += ('From: ' + self.from_ + '\n')
        if self.subject:
            self.message['text'] += ('Subject: ' + self.subject + '\n')

        self.message['text'] += self.body
        self.add_attachments()


    def add_attachments(self):
        """Add attachments."""
        if self.attachments:
            if not isinstance(self.attachments, list):
                self.attachments = [self.attachments]

            self.message['attachments'] = [{'image_url': url,
                                            'author_name': ''}
                                           for url in self.attachments]
            if self.params:
                for attachment in self.message['attachments']:
                    attachment.update(self.params)


    def send(self):
        """Send the message via HTTP POST."""
        self.construct_message()
        requests.post(self.url, json=self.message)
        print('Message sent...')
        self.sent_messages.append(repr(self))


    def send_async(self):
        """Send message asynchronously."""
        MESSAGELOOP.add_message(self)



"""messages.cli tests."""

import sys
from unittest.mock import patch

import pytest
import conftest

import click
from click import Context
from click.testing import CliRunner

from messages import MESSAGES

import messages.cli
import messages.api
from messages.cli import check_args
from messages.cli import check_type
from messages.cli import get_body
from messages.cli import trim_args
from messages.cli import main
from messages.exceptions import UnsupportedMessageTypeError
from messages.email_ import Email


##############################################################################
# FIXTURES
##############################################################################

class ClickConfig(Context):
    """Used to pass ctx for click applications."""
    def __init__(self): pass
    def get_usage(self): pass


@pytest.fixture()
def get_ctx():
    """Returns a mocked click.Context() object."""
    return ClickConfig()


##############################################################################
# TESTS: cli.check_args
##############################################################################

def test_check_args_withArgs(get_ctx):
    """
    GIVEN a call to messages via the CLI
    WHEN check_args is called with args given
    THEN assert no errors occur
    """
    ctx = get_ctx
    kwds = {'test': 1}
    check_args(ctx, kwds)


@patch.object(click, 'echo')
def test_check_args_withNoArgs(echo_mock, get_ctx):
    """
    GIVEN a call to messages via the CLI
    WHEN check_args is called with NO args given
    THEN assert click.echo and sys.exit are called
    """
    ctx = get_ctx
    kwds = {}
    with pytest.raises(SystemExit) as sysext:
        check_args(ctx, kwds)
        assert echo_mock.call_count == 2
        assert sysext.code == 0


##############################################################################
# TESTS: cli.check_type
##############################################################################

def test_check_type_normal():
    """
    GIVEN a call to messages via the CLI
    WHEN a valid message type is given
    THEN assert no errors occur
    """
    kwds = {'type': 'email'}
    check_type(kwds)


@patch.object(click, 'echo')
def test_check_type_NoType(echo_mock):
    """
    GIVEN a call to messages via the CLI
    WHEN a NO message type is given
    THEN assert click.echo and sys.exit are called
    """
    kwds = {'type': None}
    with pytest.raises(SystemExit) as sysext:
        check_type(kwds)
        assert echo_mock.call_count == 2
        assert sysext.code == 0


def test_check_type_badType():
    """
    GIVEN a call to messages via the CLI
    WHEN an invalid message type is given
    THEN assert UnsupportedMessageTypeError is raised
    """
    kwds = {'type': 'badType'}
    with pytest.raises(UnsupportedMessageTypeError):
        check_type(kwds)


##############################################################################
# TESTS: cli.get_body
##############################################################################

@conftest.travis
def test_get_body(tmpdir):
    """
    GIVEN a call to messages via the CLI
    WHEN a message is specified by filename
    THEN assert the file contents are read and set into the message body
        * this uses the tmpdir pytest built-in fixture to create temporary
          directory and file for the test
    """
    msg_file = tmpdir.join('msg.txt')
    msg_file.write('This is the message to send!')
    kwds = {'body': None, 'file': msg_file}
    body = get_body(kwds)
    assert kwds == {'body': 'This is the message to send!', 'file': None}


##############################################################################
# TESTS: cli.trim_args
##############################################################################

def test_trim_args_rejectedKV():
    """
    GIVEN a call to messages via the CLI
    WHEN trim_args is called on the CLI args
    THEN assert the correct args are removed
    """
    kwds = {'type': 'email', 'arg1': None, 'arg2': (), 'arg3': 'valid'}
    kwargs = trim_args(kwds)
    assert kwargs == {'arg3': 'valid'}


def test_trim_args_ListItems():
    """
    GIVEN a call to messages via the CLI
    WHEN trim_args is called on the CLI args
    THEN assert values with keys to/cc/bcc/attach are returned as a list
        instead of the tuple generated by the click package
    """
    kwds = {'to': ('me@here.com',), 'cc': ('her@there.com',),
            'bcc': ('him@there.com', 'her@there.com'), 'attach': ('file1',)}
    kwargs = trim_args(kwds)
    assert kwargs == {'to': ['me@here.com'], 'cc': ['her@there.com'],
            'bcc': ['him@there.com', 'her@there.com'], 'attach': ['file1']}


##############################################################################
# TESTS: cli.main
##############################################################################

@patch.object(messages.api, 'send')
@patch.object(messages.cli, 'get_body')
@patch.object(messages.cli, 'check_type')
@patch.object(messages.cli, 'check_args')
def test_main(args_mock, type_mock, body_mock, send_mock):
    runner = CliRunner()
    result = runner.invoke(main, ['-t', 'email'], catch_exceptions=True)
